/* 
   hotswap - Supports hotswapping peripherals in portable computers.

   Copyright (C) 2001 Tim Stadelmann

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
*/



/** HEADERS **/

/* autoconf header */
#include <config.h>

/* standard C headers */
#include <stdio.h>
#include <stdlib.h>

/* POSIX headers */
#include <mntent.h>

/* GNU headers */
#include <libintl.h>
#include <getopt.h>

/* system specific headers */
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/mount.h>
#include <linux/hdreg.h>
#include <linux/cdrom.h>

/* autoproject header */
#include "system.h"



/** MACROS **/

/* This macro finds the maximum of two numbers.  */
#define max(a,b) (((a) > (b)) ? (a) : (b)) 

/* the number of significant characters in the response to a yes/no
   question */
#define RESPONSE_LENGTH 8



/** PROTOTYPES **/

/* Declare prototypes for the functions in this file.  See the
   individual functions for explanations.  */
static void usage (int status);
static int decode_switches (int argc, char **argv);
static int probe_ide ();
static int rescan_ide ();
static int unregister_ide ();
static int check_mounted_ide ();



/** GLOBAL VARIABLES **/

/* The model string from the IDE device.  The maximal lenght of this
   string is hardcoded in the kernel, see linux/hdreg.h.  */
char ide_device_model[40];

/* whether an IDE device is configured */
int ide_device_present;

/* the name the program was run with */
char *program_name;

/* option flags and variables */

int nocheck; /* --nocheck */
int verbose; /* --verbose */

static struct option const long_options[] =
{
  {"version", no_argument, 0, 'V'},
  {"nocheck", no_argument, 0, 'n'},
  {"help", no_argument, 0, 'h'},
  {"verbose", no_argument, 0, 'v'},
  {NULL, 0, NULL, 0}
};




/** MAIN **/

int
main (int argc, char **argv)
{
  int i;
  int error;
  char response[RESPONSE_LENGTH];

  program_name = argv[0];

  /* Get the locale from the environment.  */
  setlocale (LC_ALL, "");
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);

  i = decode_switches (argc, argv);

  if (i == argc)
    {
      /* No ACTION argument is present.  */

      /* Enter interactive mode.  */
      
      printf (_("\
hotswap %s for notebook module bays using the secondary IDE bus\n\
Copyright 2001 Tim Stadelmann\n\
This program is free software, licensed under the conditions of the\n\
GNU General Public License version 2, or (at your option), any later\n\
version.\n\
"), VERSION);
      
      /* Check for an IDE device.  */
      error = probe_ide ();
      if (error)
	{
	  exit (EXIT_FAILURE);
	}
      
      /* This flag is set by probe_ide.  */
      if (ide_device_present) 
	{  
	  printf (_("\n\
The following IDE device is currently configured:\n\
%s\n\
Do you want to remove this device? "), ide_device_model);
	  
	  fgets(response, RESPONSE_LENGTH, stdin);
	  if (rpmatch (response) != 1)
	    {
	      puts (_("Aborting"));
	      exit (EXIT_FAILURE);
	    }
	  
	  /* Is the IDE device mounted?  */
	  error = check_mounted_ide ();
	  if (error != 0)
	    exit (EXIT_FAILURE);
	  
	  /* Unregister the IDE device.  */
	  error = unregister_ide ();
	  if (error)
	    exit (EXIT_FAILURE);
	  
	  puts (_("You can now remove the device from the module bay.\n"));
	}
      else /* No IDE device is present.  */
	{
	  puts (_("\
There is currently no IDE device configured.  (Floppy disk drives,\n\
batteries, and 'travel modules' are not managed by this utility.  If\n\
you want to swap such a module, you should remove it now.)\n\
"));
	}
      
      printf (_("\n\
Would you like to insert an IDE device into the module bay? "));
      
      fgets(response, RESPONSE_LENGTH, stdin);
      if (rpmatch (response) != 1)
	{
	  puts (_("Aborting"));
	  exit (EXIT_FAILURE);
	}
      
      printf (_("\
Please insert the new device into the module bay and press RETURN."));
      getc (stdin);
      
      /* Rescan the IDE bus.  */
    rescan:
      error = rescan_ide ();
      if (error)
	exit (EXIT_FAILURE);
      
      /* Check for an IDE device.  */
      error = probe_ide ();
      if (error)
	{
	  exit (EXIT_FAILURE);
	}
      
      if (ide_device_present)
	{  
	  printf (_("\
The following IDE device has been configured successfully:\n\
%s\n"), ide_device_model);
	}
      else /* No IDE device has been found.  */
	{
	  puts (_("\
No IDE device has been found in the module bay.  If you are sure that\n\
the device you want to configure connects to the IDE bus, make sure\n\
that the module is inserted completely.\n\
"));
	  
	  printf (_("Do you want to try again? "));

	  fgets(response, RESPONSE_LENGTH, stdin);
	  if (rpmatch (response) != 1)
	    {
	      puts (_("Aborting"));
	      exit (EXIT_FAILURE);
	    }
	  else
	    {
	      /* Go back and attempt to rescan the IDE bus again.  */
	      goto rescan;
	    }
	}
      
      /* Return successfully.  */
      exit (EXIT_SUCCESS);
    }
  
  else if (i == argc - 1)
    {
      /* There is an ACTION argument.  */

      if (strcmp ("probe-ide", argv[i]) == 0)
	{
	  /* Detect the IDE device.  */
	  error = probe_ide ();
	  if (error)
	    {
	      exit (EXIT_FAILURE);
	    }
	  else
	    {
	      printf ("%s\n", ide_device_model);
	      exit (EXIT_SUCCESS);
	    }
	}
      
      else if (strcmp ("mounted-ide", argv[i]) == 0)
	{
	  /* Is the IDE device mounted? */
	  error = check_mounted_ide ();
	  if (error == -1)
	    exit (EXIT_FAILURE);
	  
	  if (error == 1)
	    puts (_("yes"));
	  else
	    puts (_("no"));
	  
	  exit (EXIT_SUCCESS);
	}
      
      else if (strcmp ("unregister-ide", argv[i]) == 0)
	{
	  if (!nocheck)
	    {
	      /* Is the IDE device mounted? */
	      error = check_mounted_ide ();
	      if (error == -1)
		exit (EXIT_FAILURE);
	      
	      if (error == 1) 
		{	
		  printf ("%s: ", program_name);
		  puts (_("\
at least one filesystem on the IDE device is currently mounted and\n\
cannot be removed\
"));
		  exit (EXIT_FAILURE);
		}
	    }

	  /* Unregister the IDE device.  */
	  error = unregister_ide ();
	  if (error)
	    exit (EXIT_FAILURE);
	  else
	    exit (EXIT_SUCCESS);
	}
      
      else if (strcmp ("rescan-ide", argv[i]) == 0)
	{ 
	  if (!nocheck)
	    {
	      /* Is an IDE device already configured? */
	      error = probe_ide ();
	      if (error)
		exit (EXIT_FAILURE);
	      if (ide_device_present) 
		{
		  printf ("%s: ", program_name);
		  printf (_("\
an IDE device is already configured:\n\
%s\n\
"), ide_device_model);
		  exit (EXIT_FAILURE);
		}
	    }

	  /* Rescan the IDE bus.  */
	  error = rescan_ide ();
	  if (error)
	    exit (EXIT_FAILURE);
	  else
	    exit (EXIT_SUCCESS);
	}

      else
	{
	  printf ("%s: ", program_name);
	  printf (_("%s: unknown action\n\n"), argv[i]);
	  usage (EXIT_FAILURE);
	}
    }
  
  else /* Too many arguments are given.  */
    {
	  printf ("%s: ", program_name);
	  puts (_("too many arguments\n"));
	  usage (EXIT_FAILURE);
    }    
  
}



/** FUNCTIONS **/

/* Set all the option flags according to the switches specified.
   Return the index of the first non-option argument.  */

static int
decode_switches (int argc, char **argv)
{
  int c;

  while ((c = getopt_long (argc, argv, 
			   "V"  /* version */
			   "h"  /* help */
			   "n"  /* don't check */
			   "v", /* verbose */
			   long_options, (int *) 0)) != EOF)
    {
      switch (c)
	{
	case 'V':
	  printf (_("\
hotswap %s for notebook module bays using the secondary IDE bus\n\
Copyright 2001 Tim Stadelmann\n\
This program is free software, licensed under the conditions of the\n\
GNU General Public License version 2, or (at your option), any later\n\
version.\n\
"), VERSION);
	  exit (EXIT_SUCCESS);
	case 'h':
	  usage (EXIT_SUCCESS);
	case 'n':
	  nocheck = 1;
	  break;
	case 'v':
	  verbose = 1;
	  break;
	default:
	  usage (EXIT_FAILURE);
	}
    }

  return optind;
}


static void
usage (int status)
{
  printf (_("%s - \
Supports hotswapping peripherals in portable computers.\n"), program_name);
  printf (_("Usage: %s [OPTION]... [ACTION]\n"), program_name);
  puts (_("\
\n\
Options:\n\
  -h, --help                 display this help and exit\n\
  -n, --nocheck              do not perform sanity checks before\n\
                             attempting an action; use with care, as\n\
                             this option may cause actions to fail\n\
                             silently\n\
  -v, --verbose              print more information\n\
  -V, --version              output version information and exit\n\
\n\
If no ACTION is specified, interactive mode is entered.  In this mode\n\
the program guides through the hotswapping process as appropriate for\n\
the current configuration.\n\
\n\
ACTION can be one of:\n\
  probe-ide                  detect the configured IDE device\n\
  mounted-ide                check whether the IDE device is mounted\n\
  unregister-ide             unregister the IDE device\n\
  rescan-ide                 scan the IDE bus for a new device\n\
\n\
The program assumes the module bay is connected to the secondary IDE\n\
bus.  The Linux kernel requires hotswappable devices to be the only\n\
device on their IDE bus.\n\
\n\
Report bugs to <Tim.Stadelmann@physics.ox.ac.uk>.\
"));
  exit (status);
}




/* Detects whether /dev/hdc is configured and obtains the model
   identification string.  Returns -1 if an error has occured, zero
   otherwise.  The model is written into the global buffer
   ide_device_model; defaults for absent and unidentified devices are
   provided.  The global variable ide_device_present indicates whether
   a configured IDE device is present.  */
static int
probe_ide ()
{
  struct hd_driveid identity;
  int device;
  int error;

  if (verbose)
    {
      printf ("%s: ", program_name);
      puts (_("probing the configured IDE device"));
    }

  ide_device_present = 0;

  /* Open the device.  O_NONBLOCK prevents the open from failing if no
     medium is inserted.  */
  device = open ("/dev/hdc", O_RDONLY | O_NONBLOCK);
  if (device < 0)
    {
      /* A failure generally means that no drive is present.  */
      strncpy (ide_device_model, _("absent"), 40);
      /* This is not an error condition from the point of view of this
         program.  */
      return 0;
    }

  /* Get the drive identification.  */
  error = ioctl (device, HDIO_GET_IDENTITY, &identity);

  if ((errno == ENOMSG)
      || (!error && (identity.model[0] == '\0')))
    {
      /* Deal with IDE devices that do not implement drive
         identification.  Just in case, I've never actually seen one.  */
      strncpy (ide_device_model, _("unidentified IDE device"), 40);
      error = 0;
      ide_device_present = 1;
    }
  else if (error)
    {
      fprintf (stderr, "%s: ", program_name);
      perror (_("obtaining IDE drive indentification failed"));
    }
  else
    {
      strncpy (ide_device_model, identity.model, 40);
      ide_device_present = 1;
    }

  /* Close the device.  */
  close (device);

  return error;
}


/* Rescan the secondary IDE bus.  Returns -1 if an error has occured,
   zero otherwise.  */
static int
rescan_ide ()
{
  int device;
  /* These identify the second IDE bus.  See linux/ide.h.  */
  int args[] = {0x170, 0x376, 15};
  int error;

  if (verbose)
    {
      fprintf (stderr, "%s: ", program_name);
      puts (_("rescanning the IDE bus"));
    }

  /* The HDIO_SCAN_HWIF ioctl can be called on any IDE device.  We
     call it on /dev/hda since it is reasonably to assume that this
     device will always be available on a system using IDE.  */
  device = open ("/dev/hda", O_RDONLY);
  if (device < 0) {
    fprintf (stderr, "%s: ", program_name);
    perror (_("opening IDE device failed"));
    return -1;
  }

  /* Issue the ioctl.  */  
  error = ioctl (device, HDIO_SCAN_HWIF, args);
  if (error != 0)
    {
      if (errno == EIO) /* See ide.c in the kernel.  */
	{
	  /* No IDE device found.  */
	  error = 0;
	}
      else
	{
	  fprintf (stderr, "%s: ", program_name);
	  perror (_("rescanning the IDE bus failed"));
	}
    }
  /* Close the device.  */
  close (device);

  return error;
}


/* Unregister the device configured as MASTER on the secondary IDE
   bus.  Returns -1 if an error has occured, zero otherwise.  */
static int 
unregister_ide ()
{
  int device;
  int error;

  if (verbose)
    {
      fprintf (stderr, "%s: ", program_name);
      puts (_("unregistering the IDE device"));
    }

  /* If the device is a CDROM, it might still be playing.  Try and
     stop it in this case.  If it isn't a CDROM, this will generate an
     error which is ignored.  */

  /* Open the potential CDROM device; O_NONBLOCK prevents the open
     from failing if no medium is present. */
  device = open ("/dev/hdc", O_RDONLY | O_NONBLOCK);
  if (device < 0)
    {
      fprintf (stderr, "%s: ", program_name);
      perror (_("opening IDE device failed"));
    }
  /* Issue the ioctl.  */
  error = ioctl (device, CDROMSTOP);
  /* Close the device.  */
  close (device);
  
  /* The HDIO_UNREGISTER_HWIF ioctl can be called on any IDE
     device.  */
  device = open ("/dev/hda", O_RDONLY);
  if (device < 0)
    {
      fprintf (stderr, "%s: ", program_name);
      perror (_("opening IDE device failed"));
      return -1;
    }

  /* Issue the ioctl.  */
  error = ioctl (device, HDIO_UNREGISTER_HWIF, 1);
  if (error)
    {
      fprintf (stderr, "%s: ", program_name);
      perror (_("unregistering the IDE device failed"));
    }

  /* Close the device.  */
  close (device);

  /* Return on error.  */
  if (error)
    return error;

  /* The ioctl only returns an error if the calling process does not
     have root privileges.  Unregistering the device might still fail,
     for example if it is opened by some other process.  The only way
     to catch this is to probe for the device again.  Sigh.  */
  error = probe_ide ();
  if (error)
    {
      exit (EXIT_FAILURE);
    }

  if (ide_device_present)
    {
      fprintf (stderr, _("\
%s: unregistering the IDE device failed: device is probably busy\n\
"), program_name);
      error = -1;
    }

  return error;
}


/* Check whether the IDE device or a partition on it is mounted.
   Returns -1 for an error, 1 if at least one mount is active, and 0
   otherwise.  */
int
check_mounted_ide ()
{
  FILE *mtab;
  struct mntent *entry;
  int mounted = 0;

  if (verbose)
    {
      fprintf (stderr, "%s: ", program_name);
      puts (_("checking whether the IDE device is mounted"));
    }

  /* Open /etc/mtab.  */
  mtab = setmntent ("/etc/mtab", "r");
  
  if (mtab == NULL)
    {
      fprintf (stderr, "%s: ", program_name);
      perror (_("opening /etc/mtab failed"));
      return -1;
    }

  /* Loop through the entries in /etc/mtab.  */
  for (;;)
    {
      entry = getmntent (mtab);
      if (entry == NULL)
	break;
      if (strncmp("/dev/hdc", entry->mnt_fsname, 8) == 0)
	{
	  mounted = 1;
	  /* At least one partition is mounted.  */
	  break;
	}
    }
  
  /* Close /etc/mtab.  */
  endmntent (mtab);

  return mounted;
}
