/* 
   xhotswap - Motif frontend for the hotswap program.

   Copyright (C) 2001 Tim Stadelmann

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
*/



/** HEADERS **/

/* autoconf header */
#include <config.h>

/* standard C headers */
#include <stdio.h>
#include <stdlib.h>

/* GNU headers */
#include <libintl.h>
#include <getopt.h>

/* X headers */
#include <X11/keysym.h>

/* Motif headers */
#include <Xm/Xm.h>
#include <Xm/MwmUtil.h>
#include <Xm/MainW.h>
#include <Xm/RowColumn.h>
#include <Xm/MessageB.h>
#include <Xm/Frame.h>
#include <Xm/Form.h>
#include <Xm/LabelG.h>
#include <Xm/PushBG.h>
#include <Xm/CascadeBG.h>
#include <Xm/SeparatoG.h>

/* autoproject header */
#include "system.h"




/** MACROS **/

#ifndef HOTSWAP_PATH
#define HOTSWAP_PATH "hotswap"
#endif
#define BUFFER_LENGTH 256
#define ARGS_LENGTH 10


/** PROTOTYPES **/

/* Declare prototypes for external functions.  */

/* in display_error.c */
extern Cardinal display_error (Widget, String, ...);

/* Declare prototypes for the functions in this file.  See the
   individual functions for explanations.  */
static int decode_switches (int argc, char **argv);
static void create_widgets (Widget);
static int probe_device ();
static int check_mounted ();
static int do_remove ();
static int call_backend (String);
static void usage ();
static void quit_callback (Widget, XtPointer, XtPointer);
static void about_callback (Widget, XtPointer, XtPointer);
static void insert_callback (Widget, XtPointer, XtPointer);
static void insert_real_callback (Widget, XtPointer, XtPointer);
static void remove_callback (Widget, XtPointer, XtPointer);
static void swap_callback (Widget, XtPointer, XtPointer);




/** GLOBAL VARIABLES **/

/* dialog boxes */
static Widget confirmation_dialog, message_dialog;
static Widget error_dialog, retry_dialog, about_dialog;

/* the IDE device model string, limited to 40 bytes in hardware */
static char ide_device_model[40];

/* the name the program was run with */
static String program_name;

/* This buffer holds the output from the backend.  */
static char buffer[BUFFER_LENGTH];

/* the application context */
static XtAppContext application_context;

/* the application shell */
static Widget application_shell;

/* widgets that need to be accessed by several functions */
static Widget insert_button, remove_button, swap_button;
static Widget insert_menu_item, remove_menu_item, swap_menu_item;
static Widget device_label;

/* settings */
static struct settings_s {
  String hotswap_path;
} settings;

/* the resource list */
static XtResource resources[] =
{
  {"hotswapPath", "hotswapPath", XtRString, sizeof (String),
    XtOffsetOf (struct settings_s, hotswap_path), XtRString, HOTSWAP_PATH}
};

/* the option list */
static XrmOptionDescRec options[] =
{
  {"--hotswap-path", "hotswapPath", XrmoptionSepArg, (XtPointer) NULL}
};

static struct option const long_options[] =
{
  {"version", no_argument, 0, 'V'},
  {"help", no_argument, 0, 'h'},
  {NULL, 0, NULL, 0}
};




/** MAIN **/

int
main (int argc, char **argv)
{
  /* Declare variables for argument lists.  */
  Arg args[ARGS_LENGTH];
  Cardinal n;

  /* Get the locale from the environment.  */
  setlocale (LC_ALL, "");


  program_name = argv[0];

  /* Initialize gettext.  */
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);

  /* Initialize the intrinsics and create the application shell.  */
  n = 0;
  XtSetArg (args[n], XmNtitle, "Hotswap"); n++;
  XtSetArg (args[n], XmNallowShellResize, True); n++;
  XtSetArg (args[n], XmNmwmDecorations,
            MWM_DECOR_ALL | MWM_DECOR_RESIZEH | MWM_DECOR_MAXIMIZE); n++;
  application_shell = XtOpenApplication (&application_context, "Hotswap",
                                         options, XtNumber (options),
                                         &argc, argv,
                                         NULL, applicationShellWidgetClass,
                                         args, n);

  if (decode_switches (argc, argv) != argc)
    {
      /* Complain about extra arguments.  */
      usage (EXIT_FAILURE);
    }

  XtGetApplicationResources (application_shell, &settings, resources,
                             XtNumber (resources), NULL, 0);

  create_widgets (application_shell);

  /* Realize the widget hierarchy.  */
  XtRealizeWidget (application_shell);

  probe_device ();

  XtAppMainLoop (application_context);
}




/** FUNCTIONS **/

/* Set all the option flags according to the switches specified.
   Return the index of the first non-option argument.  */
static int
decode_switches (int argc, char **argv)
{
  int c;

  while ((c = getopt_long (argc, argv, 
			   "V"  /* version */
			   "h", /* help */
			   long_options, (int *) 0)) != EOF)
    {
      switch (c)
	{
	case 'V':
	  printf (_("\
Hotswap %s for notebook module bays using the secondary IDE bus\n\
Copyright 2001 Tim Stadelmann\n\
This program is free software, licensed under the conditions of the\n\
GNU General Public License version 2, or (at your option), any later\n\
version.\n\
"), VERSION);
	  exit (EXIT_SUCCESS);
	case 'h':
	  usage (EXIT_SUCCESS);
	default:
	  usage (EXIT_FAILURE);
	}
    }

  return optind;
}


static void
create_widgets (Widget application_shell)
{
  /* Declare variables for argument lists.  */
  Arg args[ARGS_LENGTH];
  Cardinal n;

  /* user interface widgets */
  Widget main_window, menu_bar, device_button, help_button, form;
  Widget device_menu, help_menu, separator, quit_menu_item;
  Widget about_menu_item;
  Widget device_frame, device_title;
  Widget action_frame, action_title, action_table;

  /* Points to the about string without the version number insterted.  */
  String raw_about_string;
  /* Points to the finalized about string.  */
  String about_string;

  /* Create the string for the about box.  */
  raw_about_string = _("\
Notebook Hotswap Utility (xhotswap) %s\n\
\n\
Copyright  2001 Tim Stadelmann\
");
  about_string
    = XtMalloc (strlen (raw_about_string) - 2 + strlen (VERSION) + 1);
  sprintf (about_string, raw_about_string, VERSION);


  /* Create the widget hierarchy.  */

  n = 0;
  main_window = XmCreateMainWindow (application_shell, "main_window", args, n);
  XtManageChild (main_window);


  /* Create the menu bar.  */

  n = 0;
  menu_bar = XmCreateMenuBar (main_window, "menu_bar", args, n);
  XtManageChild (menu_bar);


  /* Create the device menu.  */

  n = 0;
  device_menu = XmCreatePulldownMenu (menu_bar, "device_menu", args, n);

  n = 0;
  XtSetArg (args[n], XmNlabelString, XmStringCreateLocalized (_("Swap"))); n++;
  /* the mnemonic for the 'Swap' menu item */
  XtSetArg (args[n], XmNmnemonic, XStringToKeysym (_("S"))); n++;
  XtSetArg (args[n], XmNaccelerator, "Ctrl<Key>S"); n++;
  XtSetArg (args[n], XmNacceleratorText,
            XmStringCreateLocalized (_("Ctrl+S"))); n++;
  swap_menu_item = XmCreatePushButtonGadget (device_menu, "swap_menu_item",
                                             args, n);
  XtManageChild (swap_menu_item);
  XtAddCallback (swap_menu_item, XmNactivateCallback, swap_callback, NULL);

  n = 0;
  XtSetArg (args[n], XmNlabelString,
            XmStringCreateLocalized (_("Remove"))); n++;
  /* the mnemonic for the 'Remove' menu item */
  XtSetArg (args[n], XmNmnemonic, XStringToKeysym (_("R"))); n++;
  XtSetArg (args[n], XmNaccelerator, "Ctrl<Key>R"); n++;
  XtSetArg (args[n], XmNacceleratorText,
            XmStringCreateLocalized (_("Ctrl+R"))); n++;
  remove_menu_item
    = XmCreatePushButtonGadget (device_menu, "remove_menu_item", args, n);
  XtManageChild (remove_menu_item);
  XtAddCallback (remove_menu_item, XmNactivateCallback,
                 remove_callback, NULL);

  n = 0;
  XtSetArg (args[n], XmNlabelString,
            XmStringCreateLocalized (_("Insert"))); n++;
  /* the mnemonic for the 'Insert' menu item */
  XtSetArg (args[n], XmNmnemonic, XStringToKeysym (_("I"))); n++;
  XtSetArg (args[n], XmNaccelerator, "Ctrl<Key>I"); n++;
  XtSetArg (args[n], XmNacceleratorText,
            XmStringCreateLocalized (_("Ctrl+I"))); n++;
  insert_menu_item = XmCreatePushButtonGadget (device_menu, "insert_menu_item",
                                               args, n);
  XtManageChild (insert_menu_item);
  XtAddCallback (insert_menu_item, XmNactivateCallback, insert_callback, NULL);

  n = 0;
  separator = XmCreateSeparatorGadget (device_menu, "separator", args, n);
  XtManageChild (separator);

  n = 0;
  XtSetArg (args[n], XmNlabelString, XmStringCreateLocalized (_("Exit"))); n++;
  /* the mnemonic for the 'Exit' menu item */
  XtSetArg (args[n], XmNmnemonic, XStringToKeysym (_("x"))); n++;
  XtSetArg (args[n], XmNaccelerator, "Ctrl<Key>Q"); n++;
  XtSetArg (args[n], XmNacceleratorText,
            XmStringCreateLocalized (_("Ctrl+Q"))); n++;
  quit_menu_item = XmCreatePushButtonGadget (device_menu, "quit_menu_item",
                                             args, n);
  XtManageChild (quit_menu_item);
  XtAddCallback (quit_menu_item, XmNactivateCallback, quit_callback, NULL);

  /* Create the help menu.  */

  n = 0;
  help_menu = XmCreatePulldownMenu (menu_bar, "help_menu", args, n);

  n = 0;
  XtSetArg (args[n], XmNlabelString,
            XmStringCreateLocalized (_("Product Information..."))); n++;
  /* the mnemonic for the 'Product Information...' menu item */
  XtSetArg (args[n], XmNmnemonic, XStringToKeysym (_("P"))); n++;
  about_menu_item
    = XmCreatePushButtonGadget (help_menu, "about_menu_item", args, n);
  XtManageChild (about_menu_item);
  XtAddCallback (about_menu_item, XmNactivateCallback, about_callback, NULL);


  /* Create the menu bar buttons.  */
  
  n = 0;
  XtSetArg (args[n], XmNlabelString,
            XmStringCreateLocalized (_("Device"))); n++;
  /* the mnemonic for the 'Device' menu item */
  XtSetArg (args[n], XmNmnemonic, XStringToKeysym (_("D"))); n++;
  XtSetArg (args[n], XmNsubMenuId, device_menu); n++;
  device_button = XmCreateCascadeButtonGadget (menu_bar, "device_button",
                                               args, n); n++;
  XtManageChild (device_button);

  n = 0;
  XtSetArg (args[n], XmNlabelString,
            XmStringCreateLocalized (_("Help"))); n++;
  XtSetArg (args[n], XmNmnemonic, XK_H); n++;
  XtSetArg (args[n], XmNsubMenuId, help_menu); n++;
  help_button = XmCreateCascadeButtonGadget (menu_bar, "help_button",
                                             args, n); n++;
  XtManageChild (help_button);

  n = 0;
  XtSetArg (args[n], XmNmenuHelpWidget, help_button); n++;
  XtSetValues (menu_bar, args, n);



  /* Create the work area widget and its children.  */

  n = 0;
  XtSetArg (args[n], XmNhorizontalSpacing, 10); n++;
  XtSetArg (args[n], XmNverticalSpacing, 10); n++;
  XtSetArg (args[n], XmNshadowThickness, 1); n++;
  XtSetArg (args[n], XmNresizePolicy, XmRESIZE_GROW); n++;
  form = XmCreateForm (main_window, "form", args, n);
  XtManageChild (form);

  n = 0;
  XtSetArg (args[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
  device_frame = XmCreateFrame (form, "device_frame", args, n);
  XtManageChild (device_frame);

  n = 0;
  XtSetArg (args[n], XmNlabelString,
            XmStringCreateLocalized (_("Configured Device"))); n++;
  XtSetArg (args[n], XmNchildType, XmFRAME_TITLE_CHILD); n++;
  device_title = XmCreateLabelGadget (device_frame, "device_title", args, n);
  XtManageChild (device_title);

  n = 0;
  XtSetArg (args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
  XtSetArg (args[n], XmNshadowThickness, 2); n++;
  XtSetArg (args[n], XmNmarginWidth, 3); n++;
  XtSetArg (args[n], XmNmarginHeight, 3); n++;
  device_label = XmCreateLabelGadget (device_frame, "device_label", args, n);
  XtManageChild (device_label);

  n = 0;
  XtSetArg (args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg (args[n], XmNtopWidget, device_frame); n++;
  XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
  action_frame = XmCreateFrame (form, "action_frame", args, n);
  XtManageChild (action_frame);

  n = 0;
  XtSetArg (args[n], XmNlabelString,
            XmStringCreateLocalized (_("Action"))); n++;
  XtSetArg (args[n], XmNchildType, XmFRAME_TITLE_CHILD); n++;
  action_title = XmCreateLabelGadget (action_frame, "action_title", args, n);
  XtManageChild (action_title);

  n = 0;
  XtSetArg (args[n], XmNentryAlignment, XmALIGNMENT_CENTER); n++;
  XtSetArg (args[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
  action_table = XmCreateRowColumn (action_frame, "action_table", args, n);
  XtManageChild (action_table);

  n = 0;
  XtSetArg (args[n], XmNlabelString,
            XmStringCreateLocalized (_("Swap"))); n++;
  swap_button = XmCreatePushButtonGadget (action_table,
                                          "swap_button", args, n); 
  XtAddCallback (swap_button, XmNactivateCallback, swap_callback, NULL);
  XtManageChild (swap_button);

  n = 0;
  XtSetArg (args[n], XmNlabelString,
            XmStringCreateLocalized (_("Remove"))); n++;
  remove_button = XmCreatePushButtonGadget (action_table,
                                            "remove_button", args, n);
  XtAddCallback (remove_button, XmNactivateCallback, remove_callback, NULL);
  XtManageChild (remove_button);

  n = 0;
  XtSetArg (args[n], XmNlabelString,
            XmStringCreateLocalized (_("Insert"))); n++;
  insert_button
    = XmCreatePushButtonGadget (action_table, "insert_button", args, n);
  XtAddCallback (insert_button, XmNactivateCallback, insert_callback, NULL);
  XtManageChild (insert_button);

  /* Create the about box.  */
  n = 0;
  XtSetArg (args[n], XmNdialogTitle,
            XmStringCreateLocalized (_("Hotswap: Product Information"))); n++;
  XtSetArg (args[n], XmNmwmDecorations,
	    MWM_DECOR_ALL | MWM_DECOR_RESIZEH); n++;
  XtSetArg (args[n], XmNmessageString,
            XmStringCreateLocalized (about_string)); n++;
  XtSetArg (args[n], XmNokLabelString,
            XmStringCreateLocalized (_("OK"))); n++;
  XtSetArg (args[n], XmNcancelLabelString,
            XmStringCreateLocalized (_("Cancel"))); n++;
  about_dialog
    = XmCreateMessageDialog (application_shell, "about_dialog", args, n);
  XtUnmanageChild (XtNameToWidget (about_dialog, "Cancel"));
  XtUnmanageChild (XtNameToWidget (about_dialog, "Help"));

  /* Free the memory allocated for the about string.  */
  XtFree (about_string);

  /* Create the message dialog.  */
  n = 0;
  XtSetArg (args[n], XmNdialogTitle,
            XmStringCreateLocalized (_("Hotswap: Confirmation"))); n++;
  XtSetArg (args[n], XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL); n++;
  XtSetArg (args[n], XmNmwmDecorations,
	    MWM_DECOR_ALL | MWM_DECOR_RESIZEH); n++;
  XtSetArg (args[n], XmNokLabelString,
            XmStringCreateLocalized (_("OK"))); n++;
  XtSetArg (args[n], XmNcancelLabelString,
            XmStringCreateLocalized (_("Cancel"))); n++;
  confirmation_dialog
    = XmCreateMessageDialog (application_shell, "confirmation_dialog",
                             args, n);
  XtUnmanageChild (XtNameToWidget (confirmation_dialog, "Help"));
  XtAddCallback (confirmation_dialog,
                 XmNokCallback, insert_real_callback, NULL);

  /* Create the confirmation dialog.  */
  n = 0;
  XtSetArg (args[n], XmNdialogTitle,
            XmStringCreateLocalized (_("Hotswap: Message"))); n++; 
  XtSetArg (args[n], XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL); n++;
  XtSetArg (args[n], XmNmwmDecorations,
	    MWM_DECOR_ALL | MWM_DECOR_RESIZEH); n++;
  XtSetArg (args[n], XmNokLabelString,
            XmStringCreateLocalized (_("OK"))); n++;
  XtSetArg (args[n], XmNcancelLabelString,
            XmStringCreateLocalized (_("Cancel"))); n++;
  message_dialog
    = XmCreateMessageDialog (application_shell, "message_dialog",
                             args, n);
  XtUnmanageChild (XtNameToWidget (message_dialog, "Cancel"));
  XtUnmanageChild (XtNameToWidget (message_dialog, "Help"));

  
  /* Create the error dialog.  */
  n = 0;
  XtSetArg (args[n], XmNdialogTitle,
	    XmStringCreateLocalized (_("Hotswap: Error"))); n++; 
  XtSetArg (args[n], XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL); n++;
  XtSetArg (args[n], XmNmwmDecorations,
	    MWM_DECOR_ALL | MWM_DECOR_RESIZEH); n++;
  XtSetArg (args[n], XmNokLabelString,
	    XmStringCreateLocalized (_("OK"))); n++;
  error_dialog
    = XmCreateErrorDialog (application_shell, "error_dialog",
			   args, n);
  XtUnmanageChild (XtNameToWidget (error_dialog, "Cancel"));
  XtUnmanageChild (XtNameToWidget (error_dialog, "Help"));


  /* Create the retry dialog.  */
  n = 0;
  XtSetArg (args[n], XmNdialogTitle,
            XmStringCreateLocalized (_("Hotswap: Question"))); n++; 
  XtSetArg (args[n], XmNmessageString, XmStringCreateLocalized (_("\
No IDE device has been found in the\n\
module bay.  If you are sure that the\n\
device you want to configure connects to\n\
the IDE bus, verify that the module is\n\
inserted completely.\n\
\n\
Do you want to try again?\
"))); n++;
  XtSetArg (args[n], XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL); n++;
  XtSetArg (args[n], XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL); n++;
  XtSetArg (args[n], XmNmwmDecorations,
	    MWM_DECOR_ALL | MWM_DECOR_RESIZEH); n++;
  XtSetArg (args[n], XmNokLabelString,
            XmStringCreateLocalized (_("Yes"))); n++;
  XtSetArg (args[n], XmNcancelLabelString,
            XmStringCreateLocalized (_("No"))); n++;
  retry_dialog
    = XmCreateQuestionDialog (application_shell, "retry_dialog",
                             args, n);
  XtUnmanageChild (XtNameToWidget (retry_dialog, "Help"));
  XtAddCallback (retry_dialog, XmNokCallback, insert_real_callback, NULL);
}


/* Probe the installed device.  Returns -1 on error, 0 otherwise.  */
static int
probe_device ()
{
  /* Declare variables for argument lists.  */
  Arg args[ARGS_LENGTH];
  Cardinal n;

  int error;

  /* Call the backend with the appropriate options. */
  error = call_backend (" -n probe-ide");
  if (error)
    strncpy (ide_device_model, _("[error]"), 40);
  
  /* Only the side effect of the strtok function is used.  */
  strtok (buffer, "\n");
  strncpy (ide_device_model, buffer, 40);
 
  /* Set the device label accordingly.  */
  n = 0;
  XtSetArg (args[n], XmNlabelString,
            XmStringCreateLocalized (ide_device_model)); n++;
  XtSetValues (device_label, args, n);

  /* Deactivate inappropriate widgets depending on wheter a device is
     present or not. */
  if (strcmp (_("absent"), ide_device_model) == 0)
    { 
      XtSetSensitive (remove_button, False);
      XtSetSensitive (swap_button, False);
      XtSetSensitive (remove_menu_item, False);
      XtSetSensitive (swap_menu_item, False);
      XtSetSensitive (insert_button, True);
      XtSetSensitive (insert_menu_item, True);     
    }
  else if (strcmp (_("[error]"), ide_device_model) == 0)
    { 
      XtSetSensitive (remove_button, False);
      XtSetSensitive (swap_button, False);
      XtSetSensitive (remove_menu_item, False);
      XtSetSensitive (swap_menu_item, False);
      XtSetSensitive (insert_button, False);
      XtSetSensitive (insert_menu_item, False);     
    }
  else
    {
      XtSetSensitive (remove_button, True);
      XtSetSensitive (swap_button, True);
      XtSetSensitive (remove_menu_item, True);
      XtSetSensitive (swap_menu_item, True);
      XtSetSensitive (insert_button, False);
      XtSetSensitive (insert_menu_item, False);
    }

  return error;
}


/* Calls the backend to find out whether there are any mounted
   filesystems left on the device.  Returns -1 if an error occured, 1
   if there are mounted filesystems, and 0 otherwise.  */
static int
check_mounted () {

  /* a buffer for the command string */
  String command;

  /* whether there are mounted filesystems */
  Boolean mounted;

  int error;

  /* Check whether there are any mounted filesystems on the device.  */
  error = call_backend (" -n mounted-ide");
  if (error)
    return -1;
  
  if (strncmp (_("yes"), buffer, strlen (_("yes"))) == 0)
    {
      mounted = 1;
    }
  else
    {
      mounted = 0;
    }

  return mounted;
}


/* Call the backend to remove the device from the system configuration.
   Returns a non-zero value corresponding to the exit code of the backend
   if an error occured.  */
static int
do_remove ()
{
  /* Declare variables for argument lists.  */
  Arg args[ARGS_LENGTH];
  Cardinal n;

  int error;

  int mounted;

  mounted = check_mounted ();
  
  if (mounted == -1)
    {
      /* An error message is displayed from within check_mounted.  */
      return -1;
    }
  else if (mounted)
    {
      display_error (error_dialog, _("\
At least one filesystem on the current\n\
device is still mounted.  Unmount the\n\
mounted filesystems and try again.\
"));
      return -1;
    }
  
  /* Call the backend to remove the device.  */
  error = call_backend (" -n unregister-ide");
  if (error)
    return -1;

  return 0;
}


static void
quit_callback (Widget widget, XtPointer client_data,
               XtPointer call_data)
{
  /* Terminate the program. */
  exit (EXIT_SUCCESS);
}


/* Call the backend with arbitrary arguments and deal with errors.
   Reads both stdout and stderr of the backend into buffer.  Returns
   -1 if an error has occured internally and the return value of the
   backend otherwise.  */
static int 
call_backend (String arguments)
{
  /* Points to a buffer for the command string.  */
  String command;

  /* the output stream from the backend */
  FILE *backend_out;

  /* Stores return values possibly indication error conditions.  */
  int error;

  const char redirection[] = " 2>&1";

  /* Call the backend with the appropriate options. */

  /* Construct the command string passed to the shell.  The addition
     of unity to the length accounts for the terminating zero.  Note
     that the use of strncpy, strncat is unnecessary.  */
  command = XtMalloc (strlen (settings.hotswap_path)
                      + strlen (arguments) + strlen (redirection) + 1);
  strcpy (command, settings.hotswap_path);
  strcat (command, arguments);
  strcat (command, redirection);

  backend_out = popen (command, "r");

  /* The command string can now be discarded. */
  XtFree (command);
  
  if (backend_out == NULL)
    {
      display_error (error_dialog, _("\
An error occurred while calling the backend:\n\
%s"), strerror (errno));
      return -1;
    }

  /* Read the output of the backend.  */
  fgets (buffer, BUFFER_LENGTH, backend_out);
  
  /* Close the connection to the backend.  pclose does return the exit
     code of the program if it has succeeded.  The fact that at least
     one version of the Linux man page fails to mention this is a bug
     in the documentation, not in the library.  */
  error = pclose (backend_out);
  if (error == -1)
    {
      display_error (error_dialog, _("\
An error occurred while calling the backend:\n\
%s"), strerror (errno));
    }
  else if (error)
    {
      display_error (error_dialog, _("\
An error occured while running the backend:\n\
%s"), buffer);
    }
  
  return error;
}


static void
usage (int status)
{
  printf (_("\
Notebook Hotswap Utility (%s) - Supports hotswapping peripherals in\n\
portable computers.\n\
"), program_name);
  printf (_("Usage: %s [OPTION]...\n"), program_name);
  puts (_("\
\n\
Options:\n\
  -h, --help                 display this help and exit\n\
  -V, --version              output version information and exit\n\
      --hotswap-path PATH    look for the `hotswap' backend at the\n\
                             location given by PATH
\n\
Standard toolkit options are supported as well.\n\
\n\
The program assumes the module bay is connected to the secondary IDE\n\
bus.  The Linux kernel requires hotswappable devices to be the only\n\
device on their IDE bus.\n\
\n\
Report bugs to <Tim.Stadelmann@physics.ox.ac.uk>.\
"));
  exit (status);
}


/* Posts the about dialog box if the respective item is selected from
   the help menu.  */
static void
about_callback (Widget widget, XtPointer client_data, XtPointer call_data)
{
  /* Display the about dialog box. */
  XtManageChild (about_dialog);
}


/* Prompt for insertion of a new device.  */
static void
insert_callback (Widget widget, XtPointer client_data, XtPointer call_data)
{
  /* Declare variables for argument lists.  */
  Arg args[ARGS_LENGTH];
  Cardinal n;

  n = 0;
  XtSetArg (args[n], XmNmessageString, XmStringCreateLocalized (_("\
Please insert the new device into the\n\
module bay and click `OK'.\
"))); n++;
  XtSetValues (confirmation_dialog, args, n);

  XtManageChild (confirmation_dialog);
}


/* Call the backend to perform the actual scanning for a new device.
   Returns a nonzero value on error.  */
static void
insert_real_callback (Widget widget,
		      XtPointer client_data, XtPointer call_data)
{
  /* a buffer for the command string */
  String command;

  int error;
  
  error = call_backend (" -n rescan-ide");
  if (error)
    return;

  probe_device();

  if (strcmp (_("absent"), ide_device_model) == 0)
    XtManageChild (retry_dialog);
}


static void
remove_callback (Widget widget, XtPointer client_data, XtPointer call_data)
{
  /* Declare variables for argument lists.  */
  Arg args[ARGS_LENGTH];
  Cardinal n;

  /* Remove the IDE device from the system configuration.  */
  if (do_remove () != 0)
    return;

  n = 0;
  XtSetArg (args[n], XmNmessageString, XmStringCreateLocalized (_("\
You can now remove the device from the\n\
module bay.\
"))); n++;
  XtSetValues (message_dialog, args, n);

  XtManageChild (message_dialog);  

  probe_device();
}


static void
swap_callback (Widget widget, XtPointer client_data, XtPointer call_data)
{
  /* Declare variables for argument lists.  */
  Arg args[ARGS_LENGTH];
  Cardinal n;

  /* Remove the IDE device from the system configuration.  */
  if (do_remove () != 0)
    return;

  n = 0;
  XtSetArg (args[n], XmNmessageString, XmStringCreateLocalized (_("\
You can now exchange the device in the\n\
module bay.\
"))); n++;
  XtSetValues (confirmation_dialog, args, n);

  XtManageChild (confirmation_dialog);  

  probe_device ();
}
